# Copyright 2020 Zeshen Xing
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import strutils, strtabs, parseutils, tables, base64


type
  BadSecretKeyError* = object of CatchableError
  EmptySecretKeyError* = object of CatchableError
  BaseType* = int | float | bool | string
  SecretKey* = distinct string
  SecretUrl* = distinct string


  Session* = object
    data: StringTableRef
    newCreated*: bool
    modified*: bool
    accessed*: bool

  FormPart* = object
    data*: OrderedTableRef[string, tuple[params: StringTableRef, body: string]]


proc initFormPart*(): FormPart {.inline.} =
  FormPart(data: newOrderedTable[string, (StringTableRef, string)]())

proc `[]`*(formPart: FormPart, key: string): tuple[params: StringTableRef,
    body: string] {.inline.} =
  formPart.data[key]

proc `[]=`*(formPart: FormPart, key: string, body: string) {.inline.} =
  formPart.data[key] = (newStringTable(mode = modeCaseSensitive), body)

proc tryParseInt(value: sink string, default: int): int {.inline.} =
  try:
    result = parseInt(value)
  except ValueError:
    result = default

proc tryParseFloat(value: sink string, default: float): float {.inline.} =
  try:
    result = parseFloat(value)
  except ValueError:
    result = default

proc tryParseBool(value: sink string, default: bool): bool {.inline.} =
  try:
    result = parseBool(value)
  except ValueError:
    result = default

proc parseValue*[T: BaseType](value: string, default: T): T {.inline.} =
  if value.len == 0:
    return default

  when T is int:
    result = tryParseInt(value, default)
  elif T is float:
    result = tryParseFloat(value, default)
  elif T is bool:
    result = tryParseBool(value, default)
  elif T is string:
    result = value

proc len*(secretKey: SecretKey): int {.inline.} =
  string(secretKey).len

proc `$`*(secretKey: SecretKey): string {.inline.} =
  ## Hide secretKey's value
  "SecretKey(********)"

proc initSession*(data: StringTableRef, newCreated = false, modified = false,
    accessed = false): Session {.inline.} =
  Session(data: data, modified: modified)

proc update*(session: var Session) {.inline.} =
  session.accessed = true
  session.modified = true

proc `[]`*(session: var Session, key: string): string {.inline.} =
  result = session.data[key]
  session.accessed = true

proc `[]=`*(session: var Session, key, value: string) {.inline.} =
  session.data[key] = value
  update(session)

proc len*(session: Session): int {.inline.} =
  session.data.len

proc getOrDefault*(session: var Session, key: string, default = ""): string {.inline.} =
  if session.data.hasKey(key):
    result = session.data[key]
  else:
    result = default
  session.accessed = true

proc del*(session: var Session, key: string) {.inline.} =
  session.data.del(key)
  update(session)

proc clear*(session: var Session) {.inline.} =
  session.data.clear(modeCaseSensitive)
  update(session)

proc `$`*(session: Session): string {.inline.} =
  $session.data

proc parseStringTable*(tabs: StringTableRef, s: string) {.inline.} =
  # """{username: flywind, password: root}"""
  # {:}
  # make sure {':', ',', '}'} notin key or value
  if s.len <= 3:
    return
  var
    pos = 0
    key, value: string
  assert(s[pos] == '{', "StringTable String starts with '{'")
  # ignore '{'
  inc(pos)
  while true:
    pos += s.parseUntil(key, ':', pos)
    # ignore ':'
    inc(pos, 2)
    if pos >= s.len:
      break

    pos += s.parseUntil(value, {',', '}'}, pos)
    # ignore ',' or '}'
    inc(pos, 2)
    tabs[key] = value
    if pos >= s.len:
      break

proc parseStringTable*(s: string): StringTableRef {.inline.} =
  result = newStringTable(mode = modeCaseSensitive)
  parseStringTable(result, s)

proc loads*(session: Session, s: string) {.inline.} =
  ## Loads session from strings.
  session.data.parseStringTable(decode(s))

proc dumps*(session: Session): string {.inline.} =
  ## Dumps session to strings.
  encode($session)
